a:hover {text-decoration:none}

.container {margin-top:20px}
.header h1 {font-size:26px;}
.header img {float:left; margin-right:30px; margin-bottom:10px}

h2.section {color:#777; border-top:1px solid #bbb; font-size:26px; margin:1em 0; padding-top:15px;}
h4.sub-param, p, ul, .container pre.prettyprint, .container pre, table.definition { margin-left:30px;}
.main ul li { margin-left:30px;}
.main p img {background:#444; padding:2px;}
.main h2 img {float: right;}
h4.sub-title, h4.sub-end {margin-top:50px; color:#797}
h4.sub-end {margin-top:0px;}
h4.sub-param {margin-top:20px; color:#888; font-size:1.2em;}
h4.lparam, h4.tparam, h4.const {margin-left:30px;}
h4.entry {margin-top:20px; font-size:15px}
table.definition .key {min-width:100px; font-weight:bold; padding-right:12px}
table.definition .key code {font-weight:normal}
table.definition {margin-left:60px;}
table.definition td {vertical-align:top; padding-top:10px}
table.definition tr:first-child td {padding-top:0}

p.summary { border:1px dashed #aaa; background:#f4f4f4; padding:10px; border-radius:8px; margin-bottom:2em;}
.container .main pre.prettyprint,
.container .main pre {border:1px solid #bbb; padding:5px; border-radius:1px; font-size:12px;}
.container .main pre a span {color:#046;}
.container .main pre a:hover span {color:#08c;}
.container .main pre.ascii {border:none; background:none; display:table; margin-left:auto; margin-right:auto;}

p code { color:#111; background:#e0e0e0; padding:2px; border:none; font-size:100%;}
p code { color:#900; background:none; padding:0; border-radius:0;}
.nav .active > a, .nav .active:hover > a { background:#ddd; color:black; }
.nav .module:hover > a { background-color:#888;}

.local-nav li:first-child a { font-weight:bold; color:black;}
.container li.module a {color:white; background:#333; border-color:#333}
li.module span {
  background-image:url(../img/glyphicons-halflings-white.png);
  background-position: -240px -93px;
  float:left;
  width:20px;
  height:20px;
}

p.todo  {background:#cc4; padding:3px; border:2px solid #880;}
p.todo span {color:white;}
p.fixme {background:#f99; padding:3px; border:2px solid #f00;}
p.fixme span {color:white;}
p.note span {
  background-image:url(../img/glyphicons-halflings.png);
  background-position: -118px 1px;
  color:transparent;
  width:20px;
  height:20px;
  position:absolute;
  top:2px;
  left:10px;
}
p.note {padding-left:40px; position:relative;}

p.warn span {
  background-image:url(../img/glyphicons-halflings.png);
  background-position: -141px -278px;
  color:transparent;
  width:20px;
  height:20px;
  position:absolute;
  top:2px;
  left:10px;
}
p.warn {padding-left:40px; position:relative;}

h4.param span, h4.function span, h4.method span { color:#999; margin-left:0}
h4 code {font-weight:normal}

.ref {position:relative;}
.ref a {position:absolute; right:15px; color:#777}
p.todo a:hover, p.fixme a:hover {color:#08c; text-decoration:none}

footer.footer { margin-top:50px; font-size:10px; text-align:center; padding:30px 50px; background:#eee; border-top:1px solid #999 }


@media (max-width: 979px) {
  .local-nav {display:block; float:none;}
  .local-nav .affix {position:relative;}
  .global-nav {display:block; float:none; clear:both;}
  .local-nav, .global-nav, .main {padding-left:10px; margin-bottom:20px;}
}

.nav .icon-chevron-right {
  float: right;
  margin-top: 2px;
  margin-right: -6px;
  margin-left:5px;
  opacity: .25;
}
